MODULE dynhpg
   !!======================================================================
   !!                       ***  MODULE  dynhpg  ***
   !! Ocean dynamics:  hydrostatic pressure gradient trend
   !!======================================================================

   !!----------------------------------------------------------------------
   !!   dyn_hpg      : update the momentum trend with the now horizontal
   !!                  gradient of the hydrostatic pressure
   !!                  default case : k-j-i loops (vector opt. available)
   !!       hpg_ctl  : initialisation and control of options
   !!       hpg_zco  : z-coordinate scheme
   !!       hpg_zps  : z-coordinate plus partial steps (interpolation)
   !!       hpg_sco  : s-coordinate (standard jacobian formulation)
   !!       hpg_hel  : s-coordinate (helsinki modification)
   !!       hpg_wdj  : s-coordinate (weighted density jacobian)
   !!       hpg_djc  : s-coordinate (Density Jacobian with Cubic polynomial)
   !!       hpg_rot  : s-coordinate (ROTated axes scheme)
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE dynhpg_jki      !
   USE phycst          ! physical constants
   USE in_out_manager  ! I/O manager
   USE trdmod          ! ocean dynamics trends 
   USE trdmod_oce      ! ocean variables trends
   USE prtctl          ! Print control
   USE lbclnk          ! lateral boundary condition 

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC dyn_hpg                ! routine called by step.F90

#if defined key_mpp_omp
   !!----------------------------------------------------------------------
   !!   'key_mpp_omp' :                                 j-k-i loop (j-slab)
   !!----------------------------------------------------------------------
   LOGICAL, PUBLIC, PARAMETER ::   lk_dynhpg_jki = .TRUE.    !: OpenMP hpg flag
   LOGICAL, PUBLIC, PARAMETER ::   lk_dynhpg     = .FALSE.   !: vector hpg flag
#else
   !!----------------------------------------------------------------------
   !!   default case :                             k-j-i loop (vector opt.)
   !!----------------------------------------------------------------------   
   LOGICAL, PUBLIC, PARAMETER ::   lk_dynhpg_jki = .FALSE.   !: OpenMP hpg flag
   LOGICAL, PUBLIC, PARAMETER ::   lk_dynhpg     = .TRUE.    !: vector hpg flag
#endif

   !! * Share module variables
   LOGICAL  ::               & !!! ** nam_dynhpg **   hpg flags
      ln_hpg_zco = .TRUE. ,  &  ! z-coordinate - full steps
      ln_hpg_zps = .FALSE.,  &  ! z-coordinate - partial steps (interpolation)
      ln_hpg_sco = .FALSE.,  &  ! s-coordinate (standard jacobian formulation)
      ln_hpg_hel = .FALSE.,  &  ! s-coordinate (helsinki modification)
      ln_hpg_wdj = .FALSE.,  &  ! s-coordinate (weighted density jacobian)
      ln_hpg_djc = .FALSE.,  &  ! s-coordinate (Density Jacobian with Cubic polynomial)
      ln_hpg_rot = .FALSE.      ! s-coordinate (ROTated axes scheme)

   REAL(wp) ::               & !!! ** nam_dynhpg ** 
      gamm = 0.e0               ! weighting coefficient

   INTEGER  ::               &  ! 
      nhpg  =  0                 ! = 0 to 6, type of pressure gradient scheme used
      !                         ! (deduced from ln_hpg_... flags)

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Header$ 
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_hpg( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_hpg  ***
      !!
      !! ** Method  :   Call the hydrostatic pressure gradient routine 
      !!      using the scheme defined in the namelist (nhpg parameter)
      !!   
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!             - Control print  (ln_ctl)
      !!
      !! History :
      !!   9.0  !  05-10  (A. Beckmann, G. Madec) various s-coordinate options
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index

      !! * local declarations
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &
         ztrdu, ztrdv                         ! 3D temporary workspace
      !!----------------------------------------------------------------------
   
      IF( kt == nit000 )   CALL hpg_ctl      ! initialisation & control of options

      ! Temporary saving of ua and va trends (l_trddyn)
      IF( l_trddyn )   THEN
         ztrdu(:,:,:) = ua(:,:,:)  
         ztrdv(:,:,:) = va(:,:,:) 
      ENDIF      

      SELECT CASE ( nhpg )      ! Hydrastatic pressure gradient computation
      CASE ( 0 )                  ! z-coordinate 
         CALL hpg_zco( kt )
      CASE ( 1 )                  ! z-coordinate plus partial steps (interpolation)
         CALL hpg_zps( kt )
      CASE ( 2 )                  ! s-coordinate (standard jacobian formulation)
         CALL hpg_sco( kt )
      CASE ( 3 )                  ! s-coordinate (helsinki modification)
         CALL hpg_hel( kt )
      CASE ( 4 )                  ! s-coordinate (weighted density jacobian)
         CALL hpg_wdj( kt )
      CASE ( 5 )                  ! s-coordinate (Density Jacobian with Cubic polynomial)
         CALL hpg_djc( kt )
      CASE ( 6 )                  ! s-coordinate (ROTated axes scheme)
         CALL hpg_rot( kt )
      CASE ( 10 )                  ! z-coordinate
         CALL hpg_zco_jki( kt )
      CASE ( 11 )                  ! z-coordinate plus partial steps (interpolation)
         CALL hpg_zps_jki( kt )
      CASE ( 12 )                  ! s-coordinate (standard jacobian formulation)
         CALL hpg_sco_jki( kt )
      END SELECT

      ! save the hydrostatic pressure gradient trends for momentum trend diagnostics
      IF( l_trddyn )   THEN
         ztrdu(:,:,:) = ua(:,:,:) - ztrdu(:,:,:)
         ztrdv(:,:,:) = va(:,:,:) - ztrdv(:,:,:)
         CALL trd_mod( ztrdu, ztrdv, jpdtdhpg, 'DYN', kt )
      ENDIF          
      
      IF(ln_ctl) THEN         ! print sum trends (used for debugging)
         CALL prt_ctl( tab3d_1=ua, clinfo1=' hpg  - Ua: ', mask1=umask, &
            &          tab3d_2=va, clinfo2=       ' Va: ', mask2=vmask, clinfo3='dyn' )
      ENDIF
      
   END SUBROUTINE dyn_hpg


   SUBROUTINE hpg_ctl
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE hpg_ctl  ***
      !!
      !! ** Purpose :   initializations for the hydrostatic pressure gradient
      !!              computation and consistency control
      !!
      !! ** Action  :   Read the namelist namdynhpg and check the consistency
      !!      with the type of vertical coordinate used (zco, zps, sco)
      !!
      !! History :
      !!   9.0  !  05-10  (A. Beckmann)  Original code
      !!----------------------------------------------------------------------
      INTEGER ::   ioptio = 0      ! temporary integer

      NAMELIST/nam_dynhpg/ ln_hpg_zco, ln_hpg_zps, ln_hpg_sco,               &
         &                 ln_hpg_hel, ln_hpg_wdj, ln_hpg_djc, ln_hpg_rot,   &
         &                 gamm
      !!----------------------------------------------------------------------

      ! Read Namelist nam_dynhpg : pressure gradient calculation options
      REWIND ( numnam )
      READ   ( numnam, nam_dynhpg )

      ! Control print
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'dyn:hpg_ctl : hydrostatic pressure gradient control'
         WRITE(numout,*) '~~~~~~~~~~~'
         WRITE(numout,*) '       Namelist nam_dynhpg : choice of hpg scheme'
         WRITE(numout,*) '          z-coord. - full steps                          ln_hpg_zco = ', ln_hpg_zco
         WRITE(numout,*) '          z-coord. - partial steps (interpolation)       ln_hpg_zps = ', ln_hpg_zps
         WRITE(numout,*) '          s-coord. (standard jacobian formulation)       ln_hpg_sco = ', ln_hpg_sco
         WRITE(numout,*) '          s-coord. (helsinki modification)               ln_hpg_hel = ', ln_hpg_hel
         WRITE(numout,*) '          s-coord. (weighted density jacobian)           ln_hpg_wdj = ', ln_hpg_wdj
         WRITE(numout,*) '          s-coord. (Density Jacobian: Cubic polynomial)  ln_hpg_djc = ', ln_hpg_djc
         WRITE(numout,*) '          s-coord. (ROTated axes scheme)                 ln_hpg_rot = ', ln_hpg_rot
         WRITE(numout,*) '          weighting coeff. (wdj scheme)                     gamm       = ', gamm
      ENDIF

      ! set nhpg from ln_hpg_... flags
      IF( ln_hpg_zco )   nhpg = 0
      IF( ln_hpg_zps )   nhpg = 1
      IF( ln_hpg_sco )   nhpg = 2
      IF( ln_hpg_hel )   nhpg = 3
      IF( ln_hpg_wdj )   nhpg = 4
      IF( ln_hpg_djc )   nhpg = 5
      IF( ln_hpg_rot )   nhpg = 6

      ! Consitency check
      ioptio = 0 
      IF( ln_hpg_zco )   ioptio = ioptio + 1
      IF( ln_hpg_zps )   ioptio = ioptio + 1
      IF( ln_hpg_sco )   ioptio = ioptio + 1
      IF( ln_hpg_hel )   ioptio = ioptio + 1
      IF( ln_hpg_wdj )   ioptio = ioptio + 1
      IF( ln_hpg_djc )   ioptio = ioptio + 1
      IF( ln_hpg_rot )   ioptio = ioptio + 1
      IF ( ioptio > 1 ) THEN
          IF(lwp) WRITE(numout,cform_err)
          IF(lwp) WRITE(numout,*) ' several hydrostatic pressure gradient options used'
          nstop = nstop + 1
      ENDIF

      IF( lk_dynhpg_jki ) THEN
         nhpg = nhpg + 10
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) '          Autotasking or OPENMP: use j-k-i loops (i.e. _jki routines)'
      ENDIF

   END SUBROUTINE hpg_ctl


   SUBROUTINE hpg_zco( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_zco  ***
      !!
      !! ** Method  :   z-coordinate case, levels are horizontal surfaces.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level:    zhpi = grav .....
      !!                zhpj = grav .....
      !!      add it to the general momentum trend (ua,va).
      !!            ua = ua - 1/e1u * zhpi
      !!            va = va - 1/e2v * zhpj
      !! 
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!
      !! History :
      !!   1.0  !  87-09  (P. Andrich, M.-A. Foujols)  Original code
      !!   5.0  !  91-11  (G. Madec)
      !!   7.0  !  96-01  (G. Madec)
      !!   8.0  !  97-05  (G. Madec)  split dynber into dynkeg and dynhpg
      !!   8.5  !  02-07  (G. Madec)  F90: Free form and module
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace
      
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index
      
      !! * local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   &   
         zcoef0, zcoef1                 ! temporary scalars
      !!----------------------------------------------------------------------
      
      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zco : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   z-coordinate case '
      ENDIF
      
      
      ! Local constant initialization 
      ! -----------------------------
      zcoef0 = - grav * 0.5

      ! Surface value
      ! -------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! hydrostatic pressure gradient
            zhpi(ji,jj,1) = zcoef1 * ( rhd(ji+1,jj,1) - rhd(ji,jj,1) ) / e1u(ji,jj)
            zhpj(ji,jj,1) = zcoef1 * ( rhd(ji,jj+1,1) - rhd(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1)
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1)
         END DO
      END DO

      ! interior value (2=<jk=<jpkm1)
      ! --------------
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! hydrostatic pressure gradient
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd(ji+1,jj,jk)+rhd(ji+1,jj,jk-1) )   &
                  &                       - ( rhd(ji  ,jj,jk)+rhd(ji  ,jj,jk-1) )  ) / e1u(ji,jj)

               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd(ji,jj+1,jk)+rhd(ji,jj+1,jk-1) )   &
                  &                       - ( rhd(ji,jj,  jk)+rhd(ji,jj  ,jk-1) )  ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk)
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk)
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_zco


   SUBROUTINE hpg_zps( kt )
      !!---------------------------------------------------------------------
      !!                 ***  ROUTINE hpg_zps  ***
      !!                    
      !! ** Method  :   z-coordinate plus partial steps case.  blahblah...
      !! 
      !! ** Action  : - Update (ua,va) with the now hydrastatic pressure trend
      !!    
      !! History :
      !!   8.5  !  02-08  (A. Bozec)  Original code
      !!   9.0  !  04-08  (G. Madec)  F90
      !!---------------------------------------------------------------------- 
      !! * modules used                                 
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace
               
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index

      !! * local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      INTEGER ::   iku, ikv             ! temporary integers
      REAL(wp) ::   &
         zcoef0, zcoef1,             &  ! temporary scalars
         zcoef2, zcoef3                 !    "         "
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zps : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   z-coordinate with partial steps'
         IF(lwp) WRITE(numout,*) '              vector optimization'
      ENDIF


      ! 0. Local constant initialization
      ! --------------------------------
      zcoef0 = - grav * 0.5

      ! 1. Surface value
      ! ----------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! hydrostatic pressure gradient
            zhpi(ji,jj,1) = zcoef1 * ( rhd(ji+1,jj  ,1) - rhd(ji,jj,1) ) / e1u(ji,jj)
            zhpj(ji,jj,1) = zcoef1 * ( rhd(ji  ,jj+1,1) - rhd(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1)
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1)
         END DO
      END DO

      ! 2. interior value (2=<jk=<jpkm1)
      ! -----------------
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! hydrostatic pressure gradient
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd(ji+1,jj,jk) + rhd(ji+1,jj,jk-1) )   &
                  &                       - ( rhd(ji  ,jj,jk) + rhd(ji  ,jj,jk-1) )  ) / e1u(ji,jj)

               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd(ji,jj+1,jk) + rhd(ji,jj+1,jk-1) )   &
                  &                       - ( rhd(ji,jj,  jk) + rhd(ji,jj  ,jk-1) )  ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk)
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk)
            END DO
         END DO
      END DO

      ! partial steps correction at the last level  (new gradient with  intgrd.F)
# if defined key_vectopt_loop
         jj = 1
         DO ji = jpi+2, jpij-jpi-1   ! vector opt. (forced unrolling)
# else
      DO jj = 2, jpjm1
         DO ji = 2, jpim1
# endif
            iku = MIN ( mbathy(ji,jj), mbathy(ji+1,jj) ) - 1
            ikv = MIN ( mbathy(ji,jj), mbathy(ji,jj+1) ) - 1
            zcoef2 = zcoef0 * MIN( fse3w(ji,jj,iku), fse3w(ji+1,jj  ,iku) )
            zcoef3 = zcoef0 * MIN( fse3w(ji,jj,ikv), fse3w(ji  ,jj+1,ikv) )
            ! on i-direction
            IF ( iku > 2 ) THEN
               ! subtract old value
               ua(ji,jj,iku) = ua(ji,jj,iku) - zhpi(ji,jj,iku)
               ! compute the new one
               zhpi (ji,jj,iku) = zhpi(ji,jj,iku-1)   &
                  + zcoef2 * ( rhd(ji+1,jj,iku-1) - rhd(ji,jj,iku-1) + gru(ji,jj) ) / e1u(ji,jj)
               ! add the new one to the general momentum trend
               ua(ji,jj,iku) = ua(ji,jj,iku) + zhpi(ji,jj,iku)
            ENDIF
            ! on j-direction
            IF ( ikv > 2 ) THEN
               ! subtract old value
               va(ji,jj,ikv) = va(ji,jj,ikv) - zhpj(ji,jj,ikv)
               ! compute the new one
               zhpj (ji,jj,ikv) = zhpj(ji,jj,ikv-1)   &
                  + zcoef3 * ( rhd(ji,jj+1,ikv-1) - rhd(ji,jj,ikv-1) + grv(ji,jj) ) / e2v(ji,jj)
               ! add the new one to the general momentum trend
               va(ji,jj,ikv) = va(ji,jj,ikv) + zhpj(ji,jj,ikv)
            ENDIF
# if ! defined key_vectopt_loop
         END DO
# endif
      END DO

   END SUBROUTINE hpg_zps


   SUBROUTINE hpg_sco( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_sco  ***
      !!
      !! ** Method  :   s-coordinate case. Jacobian scheme.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!
      !! History :
      !!   7.0  !  96-01  (G. Madec)  s-coordinates
      !!        !  97-05  (G. Madec)  split dynber into dynkeg and dynhpg
      !!   8.5  !  02-08  (G. Madec)  F90: Free form and module, vector opt.
      !!   9.0  !  04-08  (C. Talandier) New trends organization
      !!   9.0  !  05-10  (A. Beckmann) various s-coordinate options
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index

      !! * Local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   &
         zcoef0, zuap, zvap             ! temporary scalars
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_sco : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   s-coordinate case, OPA original scheme used'
      ENDIF


      ! 0. Local constant initialization
      ! --------------------------------
      zcoef0 = - grav * 0.5


      ! 1. Surface value
      ! ----------------                          
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.   
            ! hydrostatic pressure gradient along s-surfaces
            zhpi(ji,jj,1) = zcoef0 / e1u(ji,jj) * ( fse3w(ji+1,jj  ,1) * rhd(ji+1,jj  ,1)   &
               &                                  - fse3w(ji  ,jj  ,1) * rhd(ji  ,jj  ,1) )
            zhpj(ji,jj,1) = zcoef0 / e2v(ji,jj) * ( fse3w(ji  ,jj+1,1) * rhd(ji  ,jj+1,1)   &
               &                                  - fse3w(ji  ,jj  ,1) * rhd(ji  ,jj  ,1) )
            ! s-coordinate pressure gradient correction
            zuap = -zcoef0 * ( rhd   (ji+1,jj,1) + rhd   (ji,jj,1) )   &
               &           * ( fsde3w(ji+1,jj,1) - fsde3w(ji,jj,1) ) / e1u(ji,jj)
            zvap = -zcoef0 * ( rhd   (ji,jj+1,1) + rhd   (ji,jj,1) )   &
               &           * ( fsde3w(ji,jj+1,1) - fsde3w(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1) + zuap
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1) + zvap
         END DO  
      END DO   
            
               
      ! 2. interior value (2=<jk=<jpkm1)
      ! -----------------     
      DO jk = 2, jpkm1                                  
         DO jj = 2, jpjm1     
            DO ji = fs_2, fs_jpim1   ! vector opt.      
               ! hydrostatic pressure gradient along s-surfaces
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1) + zcoef0 / e1u(ji,jj)   & 
                  &           * (  fse3w(ji+1,jj,jk) * ( rhd(ji+1,jj,jk) + rhd(ji+1,jj,jk-1) )   & 
                  &              - fse3w(ji  ,jj,jk) * ( rhd(ji  ,jj,jk) + rhd(ji  ,jj,jk-1) )  )
               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1) + zcoef0 / e2v(ji,jj)   &
                  &           * (  fse3w(ji,jj+1,jk) * ( rhd(ji,jj+1,jk) + rhd(ji,jj+1,jk-1) )   &
                  &              - fse3w(ji,jj  ,jk) * ( rhd(ji,jj,  jk) + rhd(ji,jj  ,jk-1) )  )
               ! s-coordinate pressure gradient correction
               zuap = -zcoef0 * ( rhd   (ji+1,jj  ,jk) + rhd   (ji,jj,jk) )   &
                  &           * ( fsde3w(ji+1,jj  ,jk) - fsde3w(ji,jj,jk) ) / e1u(ji,jj)
               zvap = -zcoef0 * ( rhd   (ji  ,jj+1,jk) + rhd   (ji,jj,jk) )   &
                  &           * ( fsde3w(ji  ,jj+1,jk) - fsde3w(ji,jj,jk) ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk) + zuap
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk) + zvap
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_sco


   SUBROUTINE hpg_hel( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_hel  ***
      !!
      !! ** Method  :   s-coordinate case.
      !!      The now hydrostatic pressure gradient at a given level
      !!      jk is computed by taking the vertical integral of the in-situ 
      !!      density gradient along the model level from the suface to that 
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!
      !! History :
      !!   9.0  !  05-10  (A. Beckmann)  Original code
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index
      
      !! * Local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   &
         zcoef0, zuap, zvap             ! temporary scalars
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_hel : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   s-coordinate case, helsinki modified scheme'
      ENDIF


      ! 0. Local constant initialization
      ! --------------------------------
      zcoef0 = - grav * 0.5

 
      ! 1. Surface value
      ! ----------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ! hydrostatic pressure gradient along s-surfaces
            zhpi(ji,jj,1) = zcoef0 / e1u(ji,jj) * ( fse3t(ji+1,jj  ,1) * rhd(ji+1,jj  ,1)  &
               &                                  - fse3t(ji  ,jj  ,1) * rhd(ji  ,jj  ,1) )
            zhpj(ji,jj,1) = zcoef0 / e2v(ji,jj) * ( fse3t(ji  ,jj+1,1) * rhd(ji  ,jj+1,1)  &
               &                                  - fse3t(ji  ,jj  ,1) * rhd(ji  ,jj  ,1) )
            ! s-coordinate pressure gradient correction
            zuap = -zcoef0 * ( rhd   (ji+1,jj,1) + rhd   (ji,jj,1) )   &
               &           * ( fsdept(ji+1,jj,1) - fsdept(ji,jj,1) ) / e1u(ji,jj)
            zvap = -zcoef0 * ( rhd   (ji,jj+1,1) + rhd   (ji,jj,1) )   &
               &           * ( fsdept(ji,jj+1,1) - fsdept(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1) + zuap
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1) + zvap
         END DO
      END DO

      ! 2. interior value (2=<jk=<jpkm1)
      ! -----------------
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! hydrostatic pressure gradient along s-surfaces
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1) &
                  &           +  zcoef0 / e1u(ji,jj) * ( fse3t(ji+1,jj,jk  ) * rhd(ji+1,jj,jk)     &
                  &                                     -fse3t(ji  ,jj,jk  ) * rhd(ji  ,jj,jk)   ) &
                  &           +  zcoef0 / e1u(ji,jj) * ( fse3t(ji+1,jj,jk-1) * rhd(ji+1,jj,jk-1)   &
                  &                                     -fse3t(ji  ,jj,jk-1) * rhd(ji  ,jj,jk-1) )
               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1) &
                  &           +  zcoef0 / e2v(ji,jj) * ( fse3t(ji,jj+1,jk  ) * rhd(ji,jj+1,jk)   &
                  &                                     -fse3t(ji,jj  ,jk  ) * rhd(ji,jj,  jk)   ) &
                  &           +  zcoef0 / e2v(ji,jj) * ( fse3t(ji,jj+1,jk-1) * rhd(ji,jj+1,jk-1) &
                  &                                     -fse3t(ji,jj  ,jk-1) * rhd(ji,jj,  jk-1) )
               ! s-coordinate pressure gradient correction
               zuap = - zcoef0 * ( rhd   (ji+1,jj,jk) + rhd   (ji,jj,jk) )   &
                  &            * ( fsdept(ji+1,jj,jk) - fsdept(ji,jj,jk) ) / e1u(ji,jj)
               zvap = - zcoef0 * ( rhd   (ji,jj+1,jk) + rhd   (ji,jj,jk) )   &
                  &            * ( fsdept(ji,jj+1,jk) - fsdept(ji,jj,jk) ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk) + zuap
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk) + zvap
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_hel


   SUBROUTINE hpg_wdj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_wdj  ***
      !!
      !! ** Method  :   Weighted Density Jacobian (wdj) scheme (song 1998)
      !!      The weighting coefficients from the namelist parameter gamm
      !!      (alpha=0.5-gamm ; beta=1-alpha=0.5+gamm)
      !!
      !! Reference : Song, Mon. Wea. Rev., 126, 3213-3230, 1998.
      !!
      !! History :
      !!   9.0  !  05-05  (B.W. An)  Original code
      !!        !  05-10  (G. Madec) style & small optimisation
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index

      !! * Local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   &
         zcoef0, zuap, zvap,         &  ! temporary scalars
         zalph , zbeta                  !    "         "
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_wdj : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   Weighted Density Jacobian'
      ENDIF


      ! Local constant initialization
      ! -----------------------------
      zcoef0 = - grav * 0.5
      zalph  = 0.5 - gamm        ! weighting coefficients (alpha=0.5-gamm)
      zbeta  = 0.5 + gamm        !                        (beta =1-alpha=0.5+gamm)

      ! Surface value (no ponderation)
      ! -------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ! hydrostatic pressure gradient along s-surfaces
            zhpi(ji,jj,1) = zcoef0 / e1u(ji,jj) * (  fse3w(ji+1,jj  ,1) * rhd(ji+1,jj  ,1)   &
               &                                   - fse3w(ji  ,jj  ,1) * rhd(ji  ,jj  ,1)  )
            zhpj(ji,jj,1) = zcoef0 / e2v(ji,jj) * (  fse3w(ji  ,jj+1,1) * rhd(ji  ,jj+1,1)   &
               &                                   - fse3w(ji  ,jj  ,1) * rhd(ji,  jj  ,1)  )
            ! s-coordinate pressure gradient correction
            zuap = -zcoef0 * ( rhd   (ji+1,jj,1) + rhd   (ji,jj,1) )   &
               &           * ( fsde3w(ji+1,jj,1) - fsde3w(ji,jj,1) ) / e1u(ji,jj)
            zvap = -zcoef0 * ( rhd   (ji,jj+1,1) + rhd   (ji,jj,1) )   &
               &           * ( fsde3w(ji,jj+1,1) - fsde3w(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1) + zuap
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1) + zvap
         END DO
      END DO

      ! Interior value (2=<jk=<jpkm1) (weighted with zalph & zbeta)
      ! --------------
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1) + zcoef0 / e1u(ji,jj)                            &
                  &           * (   (            fsde3w(ji+1,jj,jk  ) + fsde3w(ji,jj,jk  )        &
                  &                            - fsde3w(ji+1,jj,jk-1) - fsde3w(ji,jj,jk-1)    )   &
                  &               * (  zalph * ( rhd   (ji+1,jj,jk-1) - rhd   (ji,jj,jk-1) )      &
                  &                  + zbeta * ( rhd   (ji+1,jj,jk  ) - rhd   (ji,jj,jk  ) )  )   &
                  &             -   (            rhd   (ji+1,jj,jk  ) + rhd   (ji,jj,jk  )        &
                  &                           - rhd   (ji+1,jj,jk-1) - rhd   (ji,jj,jk-1)     )   &
                  &               * (  zalph * ( fsde3w(ji+1,jj,jk-1) - fsde3w(ji,jj,jk-1) )      &
                  &                  + zbeta * ( fsde3w(ji+1,jj,jk  ) - fsde3w(ji,jj,jk  ) )  )  )
               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1) + zcoef0 / e2v(ji,jj)                            &
                  &           * (   (           fsde3w(ji,jj+1,jk  ) + fsde3w(ji,jj,jk  )         &
                  &                           - fsde3w(ji,jj+1,jk-1) - fsde3w(ji,jj,jk-1)     )   &
                  &               * (  zalph * ( rhd   (ji,jj+1,jk-1) - rhd   (ji,jj,jk-1) )      &
                  &                  + zbeta * ( rhd   (ji,jj+1,jk  ) - rhd   (ji,jj,jk  ) )  )   &
                  &             -   (            rhd   (ji,jj+1,jk  ) + rhd   (ji,jj,jk  )        &
                  &                            - rhd   (ji,jj+1,jk-1) - rhd   (ji,jj,jk-1)    )   &
                  &               * (  zalph * ( fsde3w(ji,jj+1,jk-1) - fsde3w(ji,jj,jk-1) )      &
                  &                  + zbeta * ( fsde3w(ji,jj+1,jk  ) - fsde3w(ji,jj,jk  ) )  )  )
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk)
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk)
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_wdj


   SUBROUTINE hpg_djc( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_djc  ***
      !!
      !! ** Method  :   Density Jacobian with Cubic polynomial scheme
      !! 
      !! Reference: Shchepetkin, A.F. & J.C. McWilliams, J. Geophys. Res.,
      !!            108(C3), 3090, 2003
      !! History :
      !!   9.0  !  05-05  (B.W. An)  Original code
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,  &  ! use ta as 3D workspace
         &              zhpj => sa      ! use sa as 3D workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt     ! ocean time-step index
      
      !! * Local declarations
      INTEGER ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   &
         zcoef0, z1_10, cffu, cffx,  &  ! temporary scalars
                 z1_12, cffv, cffy,  &  !    "         "
         zep   , cffw                   !    "         "
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &  ! 3D workspace
         drhox, dzx, drhou, dzu, rho_i,     &
         drhoy, dzy, drhov, dzv, rho_j,     &
         drhoz, dzz, drhow, dzw, rho_k
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_djc : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   s-coordinate case, density Jacobian with cubic polynomial scheme'
      ENDIF


      ! 0. Local constant initialization
      ! --------------------------------
      zcoef0 = - grav * 0.5
      z1_10  = 1.0 / 10.0
      z1_12  = 1.0 / 12.0

      !----------------------------------------------------------------------------------------
      !  compute and store in provisional arrays elementary vertical and horizontal differences
      !----------------------------------------------------------------------------------------

!!bug gm   Not a true bug, but... dzz=e3w  for dzx, dzy verify what it is really

      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               drhoz(ji,jj,jk) = rhd   (ji  ,jj  ,jk) - rhd   (ji,jj,jk-1)
               dzz  (ji,jj,jk) = fsde3w(ji  ,jj  ,jk) - fsde3w(ji,jj,jk-1)
               drhox(ji,jj,jk) = rhd   (ji+1,jj  ,jk) - rhd   (ji,jj,jk  )
               dzx  (ji,jj,jk) = fsde3w(ji+1,jj  ,jk) - fsde3w(ji,jj,jk  )
               drhoy(ji,jj,jk) = rhd   (ji  ,jj+1,jk) - rhd   (ji,jj,jk  )
               dzy  (ji,jj,jk) = fsde3w(ji  ,jj+1,jk) - fsde3w(ji,jj,jk  )
            END DO
         END DO
      END DO

      !-------------------------------------------------------------------------
      ! compute harmonic averages using eq. 5.18
      !-------------------------------------------------------------------------
      zep = 1.e-15

      !!bug  gm  drhoz not defined at level 1 and used (jk-1 with jk=2)
      !!bug  gm  idem for drhox, drhoy et ji=jpi and jj=jpj

      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               cffw = 2.0 * drhoz(ji  ,jj  ,jk) * drhoz(ji,jj,jk-1)

               cffu = 2.0 * drhox(ji+1,jj  ,jk) * drhox(ji,jj,jk  )
               cffx = 2.0 * dzx  (ji+1,jj  ,jk) * dzx  (ji,jj,jk  )
  
               cffv = 2.0 * drhoy(ji  ,jj+1,jk) * drhoy(ji,jj,jk  )
               cffy = 2.0 * dzy  (ji  ,jj+1,jk) * dzy  (ji,jj,jk  )

               IF( cffw > zep) THEN
                  drhow(ji,jj,jk) = 2.0 *   drhoz(ji,jj,jk) * drhoz(ji,jj,jk-1)   &
                     &                  / ( drhoz(ji,jj,jk) + drhoz(ji,jj,jk-1) )
               ELSE
                  drhow(ji,jj,jk) = 0.e0
               ENDIF

               dzw(ji,jj,jk) = 2.0 *   dzz(ji,jj,jk) * dzz(ji,jj,jk-1)   &
                  &                / ( dzz(ji,jj,jk) + dzz(ji,jj,jk-1) )

               IF( cffu > zep ) THEN
                  drhou(ji,jj,jk) = 2.0 *   drhox(ji+1,jj,jk) * drhox(ji,jj,jk)   &
                     &                  / ( drhox(ji+1,jj,jk) + drhox(ji,jj,jk) )
               ELSE
                  drhou(ji,jj,jk ) = 0.e0
               ENDIF

               IF( cffx > zep ) THEN
                  dzu(ji,jj,jk) = 2.0*dzx(ji+1,jj,jk)*dzx(ji,jj,jk)   &
                     &            /(dzx(ji+1,jj,jk)+dzx(ji,jj,jk))
               ELSE
                  dzu(ji,jj,jk) = 0.e0
               ENDIF

               IF( cffv > zep ) THEN
                  drhov(ji,jj,jk) = 2.0 *   drhoy(ji,jj+1,jk) * drhoy(ji,jj,jk)   &
                     &                  / ( drhoy(ji,jj+1,jk) + drhoy(ji,jj,jk) )
               ELSE
                  drhov(ji,jj,jk) = 0.e0
               ENDIF

               IF( cffy > zep ) THEN
                  dzv(ji,jj,jk) = 2.0 *   dzy(ji,jj+1,jk) * dzy(ji,jj,jk)   &
                     &                / ( dzy(ji,jj+1,jk) + dzy(ji,jj,jk) )
               ELSE
                  dzv(ji,jj,jk) = 0.e0
               ENDIF

            END DO
         END DO
      END DO

      !----------------------------------------------------------------------------------
      ! apply boundary conditions at top and bottom using 5.36-5.37
      !----------------------------------------------------------------------------------
      drhow(:,:, 1 ) = 1.5 * ( drhoz(:,:, 2 ) - drhoz(:,:,  1  ) ) - 0.5 * drhow(:,:,  2  )
      drhou(:,:, 1 ) = 1.5 * ( drhox(:,:, 2 ) - drhox(:,:,  1  ) ) - 0.5 * drhou(:,:,  2  )
      drhov(:,:, 1 ) = 1.5 * ( drhoy(:,:, 2 ) - drhoy(:,:,  1  ) ) - 0.5 * drhov(:,:,  2  )

      drhow(:,:,jpk) = 1.5 * ( drhoz(:,:,jpk) - drhoz(:,:,jpkm1) ) - 0.5 * drhow(:,:,jpkm1)
      drhou(:,:,jpk) = 1.5 * ( drhox(:,:,jpk) - drhox(:,:,jpkm1) ) - 0.5 * drhou(:,:,jpkm1)
      drhov(:,:,jpk) = 1.5 * ( drhoy(:,:,jpk) - drhoy(:,:,jpkm1) ) - 0.5 * drhov(:,:,jpkm1)


      !--------------------------------------------------------------
      ! Upper half of top-most grid box, compute and store
      !-------------------------------------------------------------

!!bug gm   :  e3w-de3w = 0.5*e3w  ....  and de3w(2)-de3w(1)=e3w(2) ....   to be verified
!          true if de3w is really defined as the sum of the e3w scale factors as, it seems to me, it should be

      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            rho_k(ji,jj,1) = -grav * ( fse3w(ji,jj,1) - fsde3w(ji,jj,1) )            &
               &                   * (  rhd(ji,jj,1)                                 &
               &                     + 0.5 * ( rhd(ji,jj,2) - rhd(ji,jj,1) )         &
               &                           * ( fse3w (ji,jj,1) - fsde3w(ji,jj,1) )   &
               &                           / ( fsde3w(ji,jj,2) - fsde3w(ji,jj,1) )  ) 
         END DO
      END DO

!!bug gm    : here also, simplification is possible
!!bug gm    : optimisation: 1/10 and 1/12 the division should be done before the loop

      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.

               rho_k(ji,jj,jk) = zcoef0 * ( rhd   (ji,jj,jk) + rhd   (ji,jj,jk-1) )                                   &
                  &                     * ( fsde3w(ji,jj,jk) - fsde3w(ji,jj,jk-1) )                                   &
                  &            - grav * z1_10 * (                                                                     &
                  &     ( drhow (ji,jj,jk) - drhow (ji,jj,jk-1) )                                                     &
                  &   * ( fsde3w(ji,jj,jk) - fsde3w(ji,jj,jk-1) - z1_12 * ( dzw  (ji,jj,jk) + dzw  (ji,jj,jk-1) ) )   &
                  &   - ( dzw   (ji,jj,jk) - dzw   (ji,jj,jk-1) )                                                     &
                  &   * ( rhd   (ji,jj,jk) - rhd   (ji,jj,jk-1) - z1_12 * ( drhow(ji,jj,jk) + drhow(ji,jj,jk-1) ) )   &
                  &                             )

               rho_i(ji,jj,jk) = zcoef0 * ( rhd   (ji+1,jj,jk) + rhd   (ji,jj,jk) )                                   &
                  &                     * ( fsde3w(ji+1,jj,jk) - fsde3w(ji,jj,jk) )                                   &
                  &            - grav* z1_10 * (                                                                      &
                  &     ( drhou (ji+1,jj,jk) - drhou (ji,jj,jk) )                                                     &
                  &   * ( fsde3w(ji+1,jj,jk) - fsde3w(ji,jj,jk) - z1_12 * ( dzu  (ji+1,jj,jk) + dzu  (ji,jj,jk) ) )   &
                  &   - ( dzu   (ji+1,jj,jk) - dzu   (ji,jj,jk) )                                                     &
                  &   * ( rhd   (ji+1,jj,jk) - rhd   (ji,jj,jk) - z1_12 * ( drhou(ji+1,jj,jk) + drhou(ji,jj,jk) ) )   &
                  &                            )

               rho_j(ji,jj,jk) = zcoef0 * ( rhd   (ji,jj+1,jk) + rhd   (ji,jj,jk) )                                   &
                  &                     * ( fsde3w(ji,jj+1,jk) - fsde3w(ji,jj,jk) )                                   &
                  &            - grav* z1_10 * (                                                                      &
                  &     ( drhov (ji,jj+1,jk) - drhov (ji,jj,jk) )                                                     &
                  &   * ( fsde3w(ji,jj+1,jk) - fsde3w(ji,jj,jk) - z1_12 * ( dzv  (ji,jj+1,jk) + dzv  (ji,jj,jk) ) )   &
                  &   - ( dzv   (ji,jj+1,jk) - dzv   (ji,jj,jk) )                                                     &
                  &   * ( rhd   (ji,jj+1,jk) - rhd   (ji,jj,jk) - z1_12 * ( drhov(ji,jj+1,jk) + drhov(ji,jj,jk) ) )   &
                  &                            )

            END DO
         END DO
      END DO
      CALL lbc_lnk(rho_k,'W',1.)
      CALL lbc_lnk(rho_i,'U',1.)
      CALL lbc_lnk(rho_j,'V',1.)


      ! ---------------
      !  Surface value
      ! ---------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zhpi(ji,jj,1) = ( rho_k(ji+1,jj  ,1) - rho_k(ji,jj,1) - rho_i(ji,jj,1) ) / e1u(ji,jj)
            zhpj(ji,jj,1) = ( rho_k(ji  ,jj+1,1) - rho_k(ji,jj,1) - rho_j(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zhpi(ji,jj,1)
            va(ji,jj,1) = va(ji,jj,1) + zhpj(ji,jj,1)
         END DO
      END DO

      ! ----------------
      !  interior value   (2=<jk=<jpkm1)
      ! ----------------
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1 
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! hydrostatic pressure gradient along s-surfaces
               zhpi(ji,jj,jk) = zhpi(ji,jj,jk-1)                                &
                  &           + (  ( rho_k(ji+1,jj,jk) - rho_k(ji,jj,jk  ) )    &
                  &              - ( rho_i(ji  ,jj,jk) - rho_i(ji,jj,jk-1) )  ) / e1u(ji,jj)
               zhpj(ji,jj,jk) = zhpj(ji,jj,jk-1)                                &
                  &           + (  ( rho_k(ji,jj+1,jk) - rho_k(ji,jj,jk  ) )    &
                  &               -( rho_j(ji,jj  ,jk) - rho_j(ji,jj,jk-1) )  ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zhpi(ji,jj,jk)
               va(ji,jj,jk) = va(ji,jj,jk) + zhpj(ji,jj,jk)
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_djc


   SUBROUTINE hpg_rot( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_rot  ***
      !!
      !! ** Method  :   rotated axes scheme (Thiem and Berntsen 2005)
      !!
      !! Reference: Thiem & Berntsen, Ocean Modelling, In press, 2005.
      !! History :
      !!   9.0  !  05-07 (B.W. An)
      !!   9.0  !  05-10 (A. Beckmann) adapted to non-equidistant and masked grids
      !!----------------------------------------------------------------------
      !! * modules used
      USE oce, ONLY :   zhpi => ta,         &  ! use ta as 3D workspace
         &              zhpj => sa             ! use sa as 3D workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt            ! ocean time-step index
      
      !! * Local declarations
      INTEGER ::   ji, jj, jk                  ! dummy loop indices
      REAL(wp) ::   &
         zforg, zcoef0, zuap, zmskd1, zmskd1m,                             &
         zfrot        , zvap, zmskd2, zmskd2m
      REAL(wp), DIMENSION(jpi,jpj) ::       &  ! 2D temporary workspace
         zdistr, zsina, zcosa             
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &  ! 3D temporary workspace
         zhpiorg, zhpirot, zhpitra, zhpine, & 
         zhpjorg, zhpjrot, zhpjtra, zhpjne
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_rot : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   s-coordinate case, rotated axes scheme used'
      ENDIF

      ! -------------------------------
      !  Local constant initialization
      ! -------------------------------
      zcoef0 = - grav * 0.5
      zforg  = 0.95e0
      zfrot  = 1.e0 - zforg

      ! inverse of the distance between 2 diagonal T-points (defined at F-point) (here zcoef0/distance)
      zdistr(:,:) = zcoef0 / SQRT( e1f(:,:)*e1f(:,:) + e2f(:,:)*e1f(:,:) )

      ! sinus and cosinus of diagonal angle at F-point
      zsina(:,:) = ATAN2( e2f(:,:), e1f(:,:) )
      zcosa(:,:) = COS( zsina(:,:) )
      zsina(:,:) = SIN( zsina(:,:) )

      ! ---------------
      !  Surface value
      ! ---------------
      ! compute and add to the general trend the pressure gradients along the axes
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ! hydrostatic pressure gradient along s-surfaces
            zhpiorg(ji,jj,1) = zcoef0 / e1u(ji,jj) * (  fse3t(ji+1,jj,1) * rhd(ji+1,jj,1)   &
               &                                      - fse3t(ji  ,jj,1) * rhd(ji  ,jj,1)  )
            zhpjorg(ji,jj,1) = zcoef0 / e2v(ji,jj) * (  fse3t(ji,jj+1,1) * rhd(ji,jj+1,1)   &
               &                                      - fse3t(ji,jj  ,1) * rhd(ji,jj  ,1)  )
            ! s-coordinate pressure gradient correction
            zuap = -zcoef0 * ( rhd   (ji+1,jj  ,1) + rhd   (ji,jj,1) )   &
               &           * ( fsdept(ji+1,jj  ,1) - fsdept(ji,jj,1) ) / e1u(ji,jj)
            zvap = -zcoef0 * ( rhd   (ji  ,jj+1,1) + rhd   (ji,jj,1) )   &
               &           * ( fsdept(ji  ,jj+1,1) - fsdept(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zforg * ( zhpiorg(ji,jj,1) + zuap )
            va(ji,jj,1) = va(ji,jj,1) + zforg * ( zhpjorg(ji,jj,1) + zvap )
         END DO
      END DO

      ! compute the pressure gradients in the diagonal directions
      DO jj = 1, jpjm1
         DO ji = 1, fs_jpim1   ! vector opt.
            zmskd1 = tmask(ji+1,jj+1,1) * tmask(ji  ,jj,1)      ! mask in the 1st diagnonal
            zmskd2 = tmask(ji  ,jj+1,1) * tmask(ji+1,jj,1)      ! mask in the 2nd diagnonal
            ! hydrostatic pressure gradient along s-surfaces
            zhpitra(ji,jj,1) = zdistr(ji,jj) * zmskd1 * (  fse3t(ji+1,jj+1,1) * rhd(ji+1,jj+1,1)   &
               &                                         - fse3t(ji  ,jj  ,1) * rhd(ji  ,jj  ,1)  )
            zhpjtra(ji,jj,1) = zdistr(ji,jj) * zmskd2 * (  fse3t(ji  ,jj+1,1) * rhd(ji  ,jj+1,1)   &
               &                                         - fse3t(ji+1,jj  ,1) * rhd(ji+1,jj  ,1)  )
            ! s-coordinate pressure gradient correction
            zuap = -zdistr(ji,jj) * zmskd1 * ( rhd   (ji+1,jj+1,1) + rhd   (ji  ,jj,1) )   &
               &                           * ( fsdept(ji+1,jj+1,1) - fsdept(ji  ,jj,1) )
            zvap = -zdistr(ji,jj) * zmskd2 * ( rhd   (ji  ,jj+1,1) + rhd   (ji+1,jj,1) )   &
               &                           * ( fsdept(ji  ,jj+1,1) - fsdept(ji+1,jj,1) )
            ! back rotation
            zhpine(ji,jj,1) = zcosa(ji,jj) * ( zhpitra(ji,jj,1) + zuap )   &
               &            - zsina(ji,jj) * ( zhpjtra(ji,jj,1) + zvap )
            zhpjne(ji,jj,1) = zsina(ji,jj) * ( zhpitra(ji,jj,1) + zuap )   &
               &            + zcosa(ji,jj) * ( zhpjtra(ji,jj,1) + zvap )
         END DO
      END DO

      ! interpolate and add to the general trend the diagonal gradient
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ! averaging
            zhpirot(ji,jj,1) = 0.5 * ( zhpine(ji,jj,1) + zhpine(ji  ,jj-1,1) )
            zhpjrot(ji,jj,1) = 0.5 * ( zhpjne(ji,jj,1) + zhpjne(ji-1,jj  ,1) )
            ! add to the general momentum trend
            ua(ji,jj,1) = ua(ji,jj,1) + zfrot * zhpirot(ji,jj,1) 
            va(ji,jj,1) = va(ji,jj,1) + zfrot * zhpjrot(ji,jj,1) 
         END DO
      END DO

      ! -----------------
      ! 2. interior value (2=<jk=<jpkm1)
      ! -----------------
      ! compute and add to the general trend the pressure gradients along the axes
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! hydrostatic pressure gradient along s-surfaces
               zhpiorg(ji,jj,jk) = zhpiorg(ji,jj,jk-1)                                                 &
                  &              +  zcoef0 / e1u(ji,jj) * (  fse3t(ji+1,jj,jk  ) * rhd(ji+1,jj,jk  )   &
                  &                                        - fse3t(ji  ,jj,jk  ) * rhd(ji  ,jj,jk  )   &
                  &                                        + fse3t(ji+1,jj,jk-1) * rhd(ji+1,jj,jk-1)   &
                  &                                        - fse3t(ji  ,jj,jk-1) * rhd(ji  ,jj,jk-1)  )
               zhpjorg(ji,jj,jk) = zhpjorg(ji,jj,jk-1)                                                 &
                  &              +  zcoef0 / e2v(ji,jj) * (  fse3t(ji,jj+1,jk  ) * rhd(ji,jj+1,jk  )   &
                  &                                        - fse3t(ji,jj  ,jk  ) * rhd(ji,jj,  jk  )   &
                  &                                        + fse3t(ji,jj+1,jk-1) * rhd(ji,jj+1,jk-1)   &
                  &                                        - fse3t(ji,jj  ,jk-1) * rhd(ji,jj,  jk-1)  )
               ! s-coordinate pressure gradient correction
               zuap = - zcoef0 * ( rhd   (ji+1,jj  ,jk) + rhd   (ji,jj,jk) )   &
                  &            * ( fsdept(ji+1,jj  ,jk) - fsdept(ji,jj,jk) ) / e1u(ji,jj)
               zvap = - zcoef0 * ( rhd   (ji  ,jj+1,jk) + rhd   (ji,jj,jk) )   &
                  &            * ( fsdept(ji  ,jj+1,jk) - fsdept(ji,jj,jk) ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zforg*( zhpiorg(ji,jj,jk) + zuap )
               va(ji,jj,jk) = va(ji,jj,jk) + zforg*( zhpjorg(ji,jj,jk) + zvap )
            END DO
         END DO
      END DO

      ! compute the pressure gradients in the diagonal directions
      DO jk = 2, jpkm1
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zmskd1  = tmask(ji+1,jj+1,jk  ) * tmask(ji  ,jj,jk  )      ! level jk   mask in the 1st diagnonal
               zmskd1m = tmask(ji+1,jj+1,jk-1) * tmask(ji  ,jj,jk-1)      ! level jk-1    "               "     
               zmskd2  = tmask(ji  ,jj+1,jk  ) * tmask(ji+1,jj,jk  )      ! level jk   mask in the 2nd diagnonal
               zmskd2m = tmask(ji  ,jj+1,jk-1) * tmask(ji+1,jj,jk-1)      ! level jk-1    "               "     
               ! hydrostatic pressure gradient along s-surfaces
               zhpitra(ji,jj,jk) = zhpitra(ji,jj,jk-1)                                                       &
                  &              + zdistr(ji,jj) * zmskd1  * ( fse3t(ji+1,jj+1,jk  ) * rhd(ji+1,jj+1,jk)     &
                  &                                           -fse3t(ji  ,jj  ,jk  ) * rhd(ji  ,jj  ,jk) )   &
                  &              + zdistr(ji,jj) * zmskd1m * ( fse3t(ji+1,jj+1,jk-1) * rhd(ji+1,jj+1,jk-1)   &
                  &                                           -fse3t(ji  ,jj  ,jk-1) * rhd(ji  ,jj  ,jk-1) )
               zhpjtra(ji,jj,jk) = zhpjtra(ji,jj,jk-1)                                                       &
                  &              + zdistr(ji,jj) * zmskd2  * ( fse3t(ji  ,jj+1,jk  ) * rhd(ji  ,jj+1,jk)     &
                  &                                           -fse3t(ji+1,jj  ,jk  ) * rhd(ji+1,jj,  jk) )   &
                  &              + zdistr(ji,jj) * zmskd2m * ( fse3t(ji  ,jj+1,jk-1) * rhd(ji  ,jj+1,jk-1)   &
                  &                                           -fse3t(ji+1,jj  ,jk-1) * rhd(ji+1,jj,  jk-1) )
               ! s-coordinate pressure gradient correction
               zuap = - zdistr(ji,jj) * zmskd1 * ( rhd   (ji+1,jj+1,jk) + rhd   (ji  ,jj,jk) )   &
                  &                            * ( fsdept(ji+1,jj+1,jk) - fsdept(ji  ,jj,jk) )
               zvap = - zdistr(ji,jj) * zmskd2 * ( rhd   (ji  ,jj+1,jk) + rhd   (ji+1,jj,jk) )   &
                  &                            * ( fsdept(ji  ,jj+1,jk) - fsdept(ji+1,jj,jk) )
               ! back rotation
               zhpine(ji,jj,jk) = zcosa(ji,jj) * ( zhpitra(ji,jj,jk) + zuap )   &
                  &             - zsina(ji,jj) * ( zhpjtra(ji,jj,jk) + zvap )
               zhpjne(ji,jj,jk) = zsina(ji,jj) * ( zhpitra(ji,jj,jk) + zuap )   &
                  &             + zcosa(ji,jj) * ( zhpjtra(ji,jj,jk) + zvap )
            END DO
         END DO
      END DO

      ! interpolate and add to the general trend
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! averaging
               zhpirot(ji,jj,jk) = 0.5 * ( zhpine(ji,jj,jk) + zhpine(ji  ,jj-1,jk) )
               zhpjrot(ji,jj,jk) = 0.5 * ( zhpjne(ji,jj,jk) + zhpjne(ji-1,jj  ,jk) )
               ! add to the general momentum trend
               ua(ji,jj,jk) = ua(ji,jj,jk) + zfrot * zhpirot(ji,jj,jk) 
               va(ji,jj,jk) = va(ji,jj,jk) + zfrot * zhpjrot(ji,jj,jk) 
            END DO
         END DO
      END DO

   END SUBROUTINE hpg_rot

   !!======================================================================
END MODULE dynhpg
